/*  
    VTun - Virtual Tunnel over TCP/IP network. 

    Copyright (C) 1998-2000  Maxim Krasnyansky <max_mk@yahoo.com>

    VTun has been derived from VPPP package by Maxim Krasnyansky. 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 */

/*
 * lfd_shaper.c,v 1.1.1.2.2.2 2002/04/25 09:19:50 bergolth Exp
 */

#include "config.h"
 
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/time.h>
#include <syslog.h>

#include "vtun.h"
#include "linkfd.h"
#include "lib.h"

/* 
 * Shaper module. 
 */

#ifdef HAVE_SHAPER 

unsigned long bytes[5+1], max_speed[5+1], sp[5+1], sp2[5+1];
struct timeval curr_time, last_time[5+1];

/* 
 * Initialization function.
 */
int get_best_fd(int next,int idle[])
{
           int i;
           /* set next fd */
           {
            int oidle, onext=next;
            oidle=idle[next];
            next++;
            if((next < 1) || (next > (vtun.svr_num))) next=1;
            while( idle[next]==-1 || (idle[next] > 1 && oidle <= 1)){
                if( onext==next ) break;
                if( idle[next] != -1) vtun_syslog(LOG_DEBUG,"skip %d idle:%d", next, idle[next]);
                next++;
                if((next < 1) || (next > vtun.svr_num)) next=1;
            }
            lfd_sn= next;
           }
           /*
           */
           return next;
           /* get speed */
           for( i=1; i <= vtun.svr_num; i++){
            if( idle[i] < 0) {
             sp[i]=sp2[i]= 0;
             continue;
            }
            lfd_sn= i;
	    sp[i]= shaper_avail();
            if( sp[i] == 0) sp2[i] = -1;
            else sp2[i]= sp[i]*1000./max_speed[i];
           }
           /* get low speed fd */
           next=0;
           do{
             next++;
             if( next > vtun.svr_num) next=1;
           } while (idle[next] < 0);

           for( i=next; i <= vtun.svr_num; i++){
              if( idle[i] < 0) continue;
              if( sp2[i] >= 0 && sp2[next] > sp2[i]) next=i;
           }
           /*
           vtun_syslog(LOG_DEBUG,"%d sp1:%d sp2:%d", next, sp2[1], sp2[2]);
           */

           return next;
}
int shaper_init(struct vtun_host *host)
{
   int i;
     /* Calculate max speed bytes/sec */
   for( i=1; i <= vtun.svr_num; i++){
     max_speed[i] = host->spd_out / 8 * 1024;
     if( i==1) max_speed[i]= 8000 / 8 * 1024;
     if( i==2) max_speed[i]= 500 / 8 * 1024;
   
     /* Compensation for delays, nanosleep and so on */ 
     /*max_speed += 400;*/
     max_speed[i] += 3000;

     bytes[i] = 0;
     
     vtun_syslog(LOG_INFO,"Traffic shaping(speed %dK) initialized.", max_speed[i] * 8 /1024);	
   }
     return 0;
}

/* Shaper counter */
int shaper_counter(int len, char *in, char **out)
{ 
     int i= lfd_sn;
     /* Just count incoming bytes */
     bytes[i] += len;

     *out = in;
     return len;
}

/* Convert tv struct to milisec */
unsigned long inline tv2ms(struct timeval tv)
{
     register unsigned long ms = (tv.tv_sec * 1000)+(tv.tv_usec / 1000); 
     return  ms ? ms : 1;
}

#ifndef timersub
/* Some includes doesn't contain this macro */
#define	timersub(a, b, result)				\
  do {							\
    (result)->tv_sec = (a)->tv_sec - (b)->tv_sec;	\
    (result)->tv_usec = (a)->tv_usec - (b)->tv_usec;	\
    if ((result)->tv_usec < 0) {			\
      --(result)->tv_sec;				\
      (result)->tv_usec += 1000000;			\
    }							\
  } while (0)
#endif

/* 
 * Main shaper function.
 * Compute current speed in bytes/sec and if it is
 * higher than maximal speed stop accepting input 
 * until the speed become lower or equal to maximal.
 */
int shaper_avail(void)
{ 
     int i=lfd_sn;
     static struct timeval tv;
     register unsigned long speed;

     /* Let me know if you have faster and better time source. */
     gettimeofday(&curr_time,NULL);

     timersub(&curr_time,&last_time[i],&tv);

     /* Calculate current speed bytes/sec. 
      * (tv2ms never returns 0) */ 	
     speed = bytes[i] * 1000 / tv2ms(tv); 
	
     if( speed > max_speed[i] ){
	/* 
	 * Sleep about 1 microsec(actual sleep might be longer). 
	 * This is actually the hack to reduce CPU usage. 
	 * Without this delay we will consume 100% CPU.
         */ 
        static struct timespec ts = {0,1000};
     	nanosleep(&ts,NULL);

	/* Don't accept input */
	return 0;
     }

     if( curr_time.tv_sec > last_time[i].tv_sec ){
        last_time[i] = curr_time;
        bytes[i] = 0;
     }

     /* Accept input */
     return  speed+1;
}

struct lfd_mod lfd_shaper = {
     "Shaper",
     shaper_init,
     shaper_counter,
     shaper_avail,
     NULL,
     NULL,
     NULL,
     NULL,
     NULL
};

#else  /* HAVE_SHAPER */

int no_shaper(struct vtun_host *host)
{
     vtun_syslog(LOG_INFO, "Traffic shaping is not supported");
     return -1;
}

struct lfd_mod lfd_shaper = {
     "Shaper",
     no_shaper, NULL, NULL, NULL, NULL, NULL, NULL, NULL
};

#endif /* HAVE_SHAPER */
